# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: labm8/internal/logging.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='labm8/internal/logging.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x1clabm8/internal/logging.proto\"\x9a\x02\n\tLogRecord\x12)\n\x05level\x18\x01 \x01(\x0e\x32\x13.LogRecord.LogLevelR\x05level\x12)\n\x11\x64\x61te_utc_epoch_ms\x18\x02 \x01(\x03R\x0e\x64\x61teUtcEpochMs\x12\x1b\n\tthread_id\x18\x03 \x01(\x03R\x08threadId\x12\x1b\n\tfile_name\x18\x04 \x01(\tR\x08\x66ileName\x12\x1f\n\x0bline_number\x18\x05 \x01(\x05R\nlineNumber\x12\x18\n\x07message\x18\x06 \x01(\tR\x07message\"B\n\x08LogLevel\x12\t\n\x05\x44\x45\x42UG\x10\x00\x12\x08\n\x04INFO\x10\x01\x12\x0b\n\x07WARNING\x10\x02\x12\t\n\x05\x45RROR\x10\x03\x12\t\n\x05\x46\x41TAL\x10\x04')
)



_LOGRECORD_LOGLEVEL = _descriptor.EnumDescriptor(
  name='LogLevel',
  full_name='LogRecord.LogLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEBUG', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INFO', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WARNING', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FATAL', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=249,
  serialized_end=315,
)
_sym_db.RegisterEnumDescriptor(_LOGRECORD_LOGLEVEL)


_LOGRECORD = _descriptor.Descriptor(
  name='LogRecord',
  full_name='LogRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='level', full_name='LogRecord.level', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='level', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='date_utc_epoch_ms', full_name='LogRecord.date_utc_epoch_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='dateUtcEpochMs', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='thread_id', full_name='LogRecord.thread_id', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='threadId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='LogRecord.file_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fileName', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='line_number', full_name='LogRecord.line_number', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='lineNumber', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='LogRecord.message', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='message', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LOGRECORD_LOGLEVEL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33,
  serialized_end=315,
)

_LOGRECORD.fields_by_name['level'].enum_type = _LOGRECORD_LOGLEVEL
_LOGRECORD_LOGLEVEL.containing_type = _LOGRECORD
DESCRIPTOR.message_types_by_name['LogRecord'] = _LOGRECORD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogRecord = _reflection.GeneratedProtocolMessageType('LogRecord', (_message.Message,), dict(
  DESCRIPTOR = _LOGRECORD,
  __module__ = 'labm8.internal.logging_pb2'
  # @@protoc_insertion_point(class_scope:LogRecord)
  ))
_sym_db.RegisterMessage(LogRecord)


# @@protoc_insertion_point(module_scope)
