STRING = """\
# File: //config.proto
# Proto: BuildInfo
seconds_since_epoch: 1571330358
host: "florence"
user: "cec"
repo_dirty: true
git_commit: "8e18722724d42babd8df70712efb05362d52a9ed"
git_branch: "wip/graph"
git_tracking: "priv/wip/graph"
git_remote_url: "git@github.com:ChrisCummins/phd-priv.git"
git_commit_author: "Chris Cummins <chrisc.101@gmail.com>"
git_commit_seconds_since_epoch: 1571330166
"""
