# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='phd',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x0c\x63onfig.proto\x12\x03phd\"\xc6\x01\n\x0cGlobalConfig\x12\x14\n\x05uname\x18\x01 \x02(\tR\x05uname\x12!\n\x0c\x63onfigure_id\x18\x02 \x02(\tR\x0b\x63onfigureId\x12\x1b\n\twith_cuda\x18\x64 \x02(\x08R\x08withCuda\x12\x32\n\x07options\x18\x03 \x02(\x0b\x32\x18.phd.GlobalConfigOptionsR\x07options\x12,\n\x05paths\x18\x04 \x02(\x0b\x32\x16.phd.GlobalConfigPathsR\x05paths\"\xb9\x01\n\x13GlobalConfigOptions\x12\x1b\n\twith_cuda\x18\x01 \x02(\x08R\x08withCuda\x12\x32\n\x15update_git_submodules\x18\x02 \x02(\x08R\x13updateGitSubmodules\x12%\n\x0esymlink_python\x18\x03 \x02(\x08R\rsymlinkPython\x12*\n\x11install_git_hooks\x18\x04 \x02(\x08R\x0finstallGitHooks\"H\n\x11GlobalConfigPaths\x12\x1b\n\trepo_root\x18\x01 \x02(\tR\x08repoRoot\x12\x16\n\x06python\x18\x02 \x02(\tR\x06python\"\xf9\x02\n\tBuildInfo\x12.\n\x13seconds_since_epoch\x18\x01 \x02(\x03R\x11secondsSinceEpoch\x12\x12\n\x04host\x18\x02 \x02(\tR\x04host\x12\x12\n\x04user\x18\x03 \x02(\tR\x04user\x12\x1d\n\nrepo_dirty\x18\x05 \x02(\x08R\trepoDirty\x12\x1d\n\ngit_commit\x18\x04 \x02(\tR\tgitCommit\x12\x1d\n\ngit_branch\x18\x06 \x01(\tR\tgitBranch\x12!\n\x0cgit_tracking\x18\x07 \x01(\tR\x0bgitTracking\x12$\n\x0egit_remote_url\x18\x08 \x01(\tR\x0cgitRemoteUrl\x12*\n\x11git_commit_author\x18\n \x01(\tR\x0fgitCommitAuthor\x12\x42\n\x1egit_commit_seconds_since_epoch\x18\x0b \x01(\x03R\x1agitCommitSecondsSinceEpoch')
)




_GLOBALCONFIG = _descriptor.Descriptor(
  name='GlobalConfig',
  full_name='phd.GlobalConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uname', full_name='phd.GlobalConfig.uname', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='uname', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='configure_id', full_name='phd.GlobalConfig.configure_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='configureId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='with_cuda', full_name='phd.GlobalConfig.with_cuda', index=2,
      number=100, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='withCuda', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='options', full_name='phd.GlobalConfig.options', index=3,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='options', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='paths', full_name='phd.GlobalConfig.paths', index=4,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='paths', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=220,
)


_GLOBALCONFIGOPTIONS = _descriptor.Descriptor(
  name='GlobalConfigOptions',
  full_name='phd.GlobalConfigOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='with_cuda', full_name='phd.GlobalConfigOptions.with_cuda', index=0,
      number=1, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='withCuda', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='update_git_submodules', full_name='phd.GlobalConfigOptions.update_git_submodules', index=1,
      number=2, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='updateGitSubmodules', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symlink_python', full_name='phd.GlobalConfigOptions.symlink_python', index=2,
      number=3, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='symlinkPython', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='install_git_hooks', full_name='phd.GlobalConfigOptions.install_git_hooks', index=3,
      number=4, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='installGitHooks', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=223,
  serialized_end=408,
)


_GLOBALCONFIGPATHS = _descriptor.Descriptor(
  name='GlobalConfigPaths',
  full_name='phd.GlobalConfigPaths',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='repo_root', full_name='phd.GlobalConfigPaths.repo_root', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='repoRoot', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='python', full_name='phd.GlobalConfigPaths.python', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='python', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=410,
  serialized_end=482,
)


_BUILDINFO = _descriptor.Descriptor(
  name='BuildInfo',
  full_name='phd.BuildInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='seconds_since_epoch', full_name='phd.BuildInfo.seconds_since_epoch', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='secondsSinceEpoch', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='host', full_name='phd.BuildInfo.host', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='host', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user', full_name='phd.BuildInfo.user', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='user', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='repo_dirty', full_name='phd.BuildInfo.repo_dirty', index=3,
      number=5, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='repoDirty', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_commit', full_name='phd.BuildInfo.git_commit', index=4,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitCommit', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_branch', full_name='phd.BuildInfo.git_branch', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitBranch', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_tracking', full_name='phd.BuildInfo.git_tracking', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitTracking', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_remote_url', full_name='phd.BuildInfo.git_remote_url', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitRemoteUrl', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_commit_author', full_name='phd.BuildInfo.git_commit_author', index=8,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitCommitAuthor', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_commit_seconds_since_epoch', full_name='phd.BuildInfo.git_commit_seconds_since_epoch', index=9,
      number=11, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitCommitSecondsSinceEpoch', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=485,
  serialized_end=862,
)

_GLOBALCONFIG.fields_by_name['options'].message_type = _GLOBALCONFIGOPTIONS
_GLOBALCONFIG.fields_by_name['paths'].message_type = _GLOBALCONFIGPATHS
DESCRIPTOR.message_types_by_name['GlobalConfig'] = _GLOBALCONFIG
DESCRIPTOR.message_types_by_name['GlobalConfigOptions'] = _GLOBALCONFIGOPTIONS
DESCRIPTOR.message_types_by_name['GlobalConfigPaths'] = _GLOBALCONFIGPATHS
DESCRIPTOR.message_types_by_name['BuildInfo'] = _BUILDINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GlobalConfig = _reflection.GeneratedProtocolMessageType('GlobalConfig', (_message.Message,), dict(
  DESCRIPTOR = _GLOBALCONFIG,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.GlobalConfig)
  ))
_sym_db.RegisterMessage(GlobalConfig)

GlobalConfigOptions = _reflection.GeneratedProtocolMessageType('GlobalConfigOptions', (_message.Message,), dict(
  DESCRIPTOR = _GLOBALCONFIGOPTIONS,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.GlobalConfigOptions)
  ))
_sym_db.RegisterMessage(GlobalConfigOptions)

GlobalConfigPaths = _reflection.GeneratedProtocolMessageType('GlobalConfigPaths', (_message.Message,), dict(
  DESCRIPTOR = _GLOBALCONFIGPATHS,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.GlobalConfigPaths)
  ))
_sym_db.RegisterMessage(GlobalConfigPaths)

BuildInfo = _reflection.GeneratedProtocolMessageType('BuildInfo', (_message.Message,), dict(
  DESCRIPTOR = _BUILDINFO,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.BuildInfo)
  ))
_sym_db.RegisterMessage(BuildInfo)


# @@protoc_insertion_point(module_scope)
